#!/usr/bin/perl -w

###
# Pod Documentation
###

=head1 NAME

CddDescriptions.pm

=head1 SYNOPSIS

=head1 DESCRIPTION

This module implements an object for reading the "cddid.tbl" file and accessing this information 
about domains from the Conserved Domain Database.

=head1 AUTHOR

Jeffrey Barrick

=head1 COPYRIGHT

Copyright (C) 2008 Jeffrey Barrick.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut

###
# End Pod Documentation
###

use Bio::Root::Root;

package CddDescriptions;
use vars qw(@ISA);
use strict;

@ISA = qw( Bio::Root::RootI );

=head2 new

 Title   : new
 Usage   : $cdd_desc = Intergenic::CddDescriptions->new();
 Function: Loads description of entries from the Conserved Domain Database
 Returns : a new CddDescriptions object

=cut

sub new
{
	my($caller,@args) = @_;
	my $class = ref($caller) || $caller;

	my $self = new Bio::Root::Root($caller, @args);

	bless ($self, $class);
	$self->{file_name} = $self->Bio::Root::RootI::_rearrange([qw(FILE_NAME)], @args);
	return 0 if (!$self->{file_name});

	$self->_load_table();
	
	return $self;
}

=head2 cdd_info

 Title   : cdd_info
 Usage   : $cdd_info = Intergenic::CddDescriptions->cdd_info( 'COG0520' );
 Function: Retrieves cdd info about the specified accession number.
 Returns : Hash to data for cdd

=cut

sub cdd_info
{
	my ($self, $cdd_accession) = @_;
	
	return $self->{table}->{$cdd_accession};
}

=head2 _load_table

 Title   : _load_table
 Usage   : Intergenic::CddDescriptions->_load_table();
 Function: Loads cdd table information .
 Returns : nothing

=cut

sub _load_table
{
	my ($self) = @_;
	$self->throw("No file name supplied for description table") if (! $self->{file_name});	
	
	open TABLE, "<$self->{file_name}";
	
	$self->{table} = {};
	while (<TABLE>)
	{
		chomp;
		my $entry;
		( $entry->{accession}, $entry->{xref}, $entry->{gene}, $entry->{desc}, $entry->{length} )
			= split /\t/, $_;
		
		$self->{table}->{$entry->{accession}} = $entry;
	}		
}

=head2 xref_to_id

 Title   : xref_to_id
 Usage   : Intergenic::CddDescriptions->xref_to_id("COG0531");
 Function: Convert an xref name into an id 
 Returns : cdd id

=cut

sub xref_to_id
{
	my ($self, $xref) = @_;
	foreach my $key (sort keys %{$self->{table}})
	{
		return $self->{table}->{$key}->{accession} if ($xref eq $self->{table}->{$key}->{xref});
	}
	
	return undef;
}

return 1;