###
# Pod Documentation
###

=head1 NAME

GenomicContext.pm

=head1 SYNOPSIS

=head1 DESCRIPTION

Object that collects information about genomic features near a hit.

=head1 AUTHOR

Jeffrey Barrick

=head1 COPYRIGHT

Copyright (C) 2008 Jeffrey Barrick.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut

###
# End Pod Documentation
###

use Bio::Root::RootI;
use Bio::DB::Flat;

use Storable;

package GenomeContext;
use vars qw(@ISA);
use strict;

@ISA = qw( Bio::Root::RootI );

our %seq_hash;
our %gene_hash;

sub new_neighborhood
{
	my($caller,@args) = @_;
	my $class = ref($caller) || $caller;
	
	my $self = Bio::Root::Root->new($caller, @args);
	bless ($self, $class);

	#Width of interval (centered on center of start-end)
	$self->{width} = $self->Bio::Root::RootI::_rearrange([qw(WIDTH)], @args);
	
	#Upstream and downstream
	($self->{upstream}, $self->{downstream}) = $self->Bio::Root::RootI::_rearrange([qw(UPSTREAM DOWNSTREAM)], @args);
	$self->warn("Both neighborhood width and upstream/downstream specified, width will be used") if (defined $self->{width} && (defined $self->{upstream} || defined $self->{upstream}));

	#Sequence location of main feature
	$self->{fragment_accession} = $self->Bio::Root::RootI::_rearrange([qw(ACCESSION)], @args);
	$self->{main}->{start} = $self->Bio::Root::RootI::_rearrange([qw(START)], @args);	
	$self->{main}->{end} = $self->Bio::Root::RootI::_rearrange([qw(END)], @args);	
	$self->{main}->{strand} = $self->Bio::Root::RootI::_rearrange([qw(STRAND)], @args);	

	($self->{db_path}) = $self->Bio::Root::RootI::_rearrange([qw(DB_PATH)], @args);	
	if ($self->{db_path})
	{
		$self->{db_path} =~ m/^(.+)\/(.+)$/;
		($self->{db_dir}, $self->{db_name}) = ($1, $2);
	}
	
	$self->{seq_hash} = $self->Bio::Root::RootI::_rearrange([qw(SEQ_HASH)], @args);
	
	$self->{record_hash} = $self->Bio::Root::RootI::_rearrange([qw(RECORD_HASH)], @args);
	
	$self->{record_cache_path} = $self->Bio::Root::RootI::_rearrange([qw(RECORD_CACHE_PATH)], @args);
	mkdir $self->{record_cache_path} if ($self->{record_cache_path} && (!-e $self->{record_cache_path}));

	$self->_fix_coords;

	#$self->{verbose} = 1;

	$self->{force_linear} = 1; #$self->Bio::Root::RootI::_rearrange([qw(FORCE_LINEAR)], @args);
	$self->_load_neighborhood();

	return $self;
}



sub _fix_coords
{
	my ($self) = @_;

	#If they did not give a strand, then calculate one and correct coords
	if (! defined $self->{main}->{strand})
	{
		$self->{main}->{strand} = + 1;
		
		if ($self->{main}->{end} < $self->{main}->{start})
		{
			$self->{main}->{strand} = -1;
			my $temp = $self->{main}->{end};
			$self->{main}->{end} = $self->{main}->{start};
			$self->{main}->{start} = $temp;
		}
	}
	
	if (!defined $self->{main}->{strand} || !defined $self->{main}->{start} || !defined $self->{main}->{end} || !defined $self->{fragment_accession})
	{
		$self->throw("$self->{fragment_accession} $self->{main}->{strand} $self->{main}->{start} $self->{main}->{end} Not valid values.")
	}
}

#Human readable/editable format input/output

sub print_feature_list
{
	my ($self, $fh) = @_;
	$fh = *STDOUT if (!$fh);
	
	print $fh +(join "\t", ($self->{fragment_accession}, $self->{genome_info}->{description}, "(1..$self->{genome_info}->{length})", $self->{genome_info}->{topology})) . "\n";
	print $fh +(join "\t", ($self->{interval}->{strand}, $self->{interval}->{start}, $self->{interval}->{end}, "Interval")) . "\n";
	print $fh +(join "\t", ($self->{strand}, $self->{start}, $self->{end}, "Hit")) . "\n";
	foreach my $gene (@{$self->{neighborhood}})
	{
		print $fh +(join "\t", ($gene->{strand}, $gene->{start}, $gene->{end}, $gene->{name}, $gene->{cog}, $gene->{description}, $gene->{cog_name}, $gene->{cog_description}, $gene->{accession} )) . "\n";
	}
}

sub _load_neighborhood
{
	my ($self) = @_;
	
	my $record = $self->_accession_to_record($self->{fragment_accession});
	$record or die "Could not load sequence in database: $self->{fragment_accession}";

	#Initialize	from record
	$self->{effective_topology} = $record->{topology}; 
	$self->{effective_topology} = 'linear' if ($self->{force_linear});
	$self->{description} = $record->{description};
	$self->{display_name} = $record->{display_name};
	$self->{length} = CalculateLength($self->{main}->{start}, $self->{main}->{end}, $record->{length}, $self->{effective_topology});
	$self->{interval}->{strand} = $self->{main}->{strand};
	$self->{taxonomy} = $record->{taxonomy};
	
	#Calculate the coordinates of the interval neighborhood	
	if (defined $self->{width})
	{
		$self->{interval}->{start} = OffsetPosition($self->{main}->{start}, -($self->{width} - $self->{length}) / 2, $record->{length}, $self->{effective_topology});
		$self->{interval}->{end} = OffsetPosition($self->{interval}->{start}, $self->{width}, $record->{length}, $self->{effective_topology});
	}
	else
	{
		$self->{interval}->{start} = OffsetPosition($self->{main}->{start}, ($self->{main}->{strand} == +1) ? -$self->{upstream} : -$self->{downstream}, $record->{length}, $self->{effective_topology});
		$self->{interval}->{end} = OffsetPosition($self->{main}->{end}, ($self->{main}->{strand} == +1) ? $self->{downstream} : $self->{upstream}, $record->{length}, $self->{effective_topology});
	}

	#Cull the gene list
	@{$self->{neighborhood}} = grep 
		{ 
			($_->{start} >= $self->{interval}->{start}) and  ($_->{start} <= $self->{interval}->{end}) 
		||	($_->{end}   >= $self->{interval}->{start}) and  ($_->{end}   <= $self->{interval}->{end}) 
		} @{$record->{gene_list}};
		
	undef $record;
}

sub _accession_to_record
{
	my ($self, $fragment_accession) = @_;

	my $record;
	
	#1: See if it already in memory 
	if (defined $self->{record_hash} and defined $self->{record_hash}->{$fragment_accession})
	{
		print "Loading $self->{fragment_accession} from memory.\n" if ($self->{verbose});
		return $self->{record_hash}->{$fragment_accession};
	}
	
	#2: See if a cache file exists
	my $cache_file; 
	if ($self->{record_cache_path})
	{
		$cache_file = "$self->{record_cache_path}/$fragment_accession";
		if (-e $cache_file)
		{
			print "Loading $self->{fragment_accession} from cache.\n" if ($self->{verbose});
			$record = Storable::retrieve($cache_file);
			$self->{record_hash}->{$fragment_accession} = $record  if (defined $self->{record_hash});

			return $record;
		}
	}
	
	#3: See if the sequence has already been loaded
	my $seq;
	
	if (defined $self->{seq_hash})
	{
		$seq = $self->{seq_hash}->{$self->{fragment_accession}};
	}
	
	#4: Load the sequence from the database
	if (!defined $seq)
	{
		print "Loading $self->{fragment_accession} from database.\n" if ($self->{verbose});
		my $database = Bio::DB::Flat->new(	-directory  => $self->{db_dir},
                        -dbname     => $self->{db_name},
                        -format     => 'genbank',
                        -index      => 'flat',
                        -write_flag => 0);
		$seq = $database->get_Seq_by_id($self->{fragment_accession});
		
		$self->throw("Could not find sequence with accession $self->{fragment_accession} in database\n") if (!$seq);
		
	}

	##save overall information 
	$record->{effective_topology} = $seq->is_circular ? 'circular' : 'linear';
	$record->{description} = $seq->description;
	$record->{display_name} = $seq->display_name;
	$record->{length} = $seq->length;
	
	my @tax_array = $seq->species->classification();
	$record->{taxonomy} = '';
	if (@tax_array)
	{	
		$record->{taxonomy} = $tax_array[scalar @tax_array - 1];
		for (my $i = scalar @tax_array - 2; $i >= 0; $i--)
		{
			$record->{taxonomy} .= "; $tax_array[$i]"; 
		}
	}
	
	##make the gene list
	my @Gene_List;	
	my @Feature_List = $seq->get_SeqFeatures(); # just top level 
	
	my $NPG_Counter = 1;
	FEATURE:  foreach my $Feature ( @Feature_List ) 
	{ 
		if ( ($Feature->primary_tag eq 'CDS') || ($Feature->primary_tag eq 'rRNA') || ($Feature->primary_tag eq 'tRNA') || ($Feature->primary_tag eq 'misc_RNA') )
		{
			
			my $Gene;
		
			$Gene->{accession} = GetTag($Feature, "protein_id");
			
			if (!$Gene->{accession})
			{
				$Gene->{accession} = "NPG_" . $NPG_Counter;
				$NPG_Counter++;
			}
			
			#$Gene->{accession} = GetTag($Feature, "locus_tag") if (!$Gene->{accession});
			#$Gene->{accession} = GetTag($Feature, "gene") if (!$Gene->{accession});

			$Gene->{name} = GetTag($Feature, "gene");
			$Gene->{name} = GetTag($Feature, "locus_tag") if (!$Gene->{name});
			$Gene->{description} = GetTag($Feature, "product");
			$Gene->{description} = '' if (!defined $Gene->{description});
#			$Gene->{aa_seq} = GetTag($Feature, "translation");
			$Gene->{locus_tag} = GetTag($Feature, "locus_tag");
			#Can't just use end and start because joins are confused...
			my @Location_List = $Feature->location->each_Location(); 
			
			#All of this information is meaningless in some contexts, like trans-spliced genes
			#so we really have to use the information from each exon instead
			$Gene->{start} = $Location_List[0]->start;
			$Gene->{end} = $Location_List[$#Location_List]->end;
			$Gene->{strand} =  $Feature->location->strand;
			$Gene->{strand} = '?' if (!defined $Gene->{strand});
			
#			#Get the GI from the db_xref field
#			my @xref_list = GetTag($Feature, "db_xref", 1);
#			XREF: foreach my $xref (@xref_list)
#			{
#				if ( $xref =~ m/^GI:(\d+)$/ )
#				{
#					$Gene->{gi} = $1;
#					last XREF;
#				}
#			}
#			
#			
#			#Create the list of exons
#			undef $Gene->{exons};
#			#my $Last_Location;
#			LOCATION: foreach my $Location (@Location_List)
#			{
#				#This happens when there are frameshifts, check for overlap to avoid
#				#False splits in gene if it overlaps the origin of a circular chromosome
#				#if ($Last_Location && ($Sequence->is_circular && ($Location->start == 1) && ($Last_Location->end == $Sequence->length)))
#				#{
#				#	#Revise current string
#				#	$Gene->{exons} =~ s/^(.+-).+?$/$1/;
#				#	$Gene->{exons} .= $Location->end;
#				#	next LOCATION;
#				#}
#			
#				my $Exon = { strand=>$Location->strand, start=>$Location->start, end=>$Location->end};
#				
#				push @{$Gene->{exons}}, $Exon;
#				
#			#	$Gene->{exons} .= ", " if ($Gene->{exons});
#			#	$Gene->{exons} .= $Location->strand . " " . $Location->start . " " . $Location->end;
#			#} continue {
#			#	$Last_Location = $Location;
#			}
			
			push @Gene_List, $Gene;
			#print Dumper($Gene);
		}
		
		#Handle other features. Right now in RefSeq the CDD are included as a /db_xref  
		elsif ($Feature->primary_tag eq 'misc_feature')
		{
			my $locus_tag = GetTag($Feature, "locus_tag");
			if ($locus_tag && defined $Gene_List[-1] && ($locus_tag eq $Gene_List[-1]->{locus_tag}))
			{
				my @db_xrefs = GetTag($Feature, "db_xref");
				DB_XREF: foreach my $db_xref (@db_xrefs)
				{
					if ($db_xref =~ m/CDD:(\d+)/)
					{
						next DB_XREF if ($Gene_List[-1]->{cdd_hash}->{$1});
						my $cdd_bounds = { start => $Feature->start, end => $Feature->end, strand=> $Feature->strand };
						push @{$Gene_List[-1]->{cdd_hash}->{$1}}, $cdd_bounds;
					}
					#cdd_gene
					#cdd_desc
				}
			}
			#Check to see if the locus_tag matches the last 
		}	
	}
	
	#Sort based on start, strand, end
	sub by_position
	{
		return ($a->{start} <=> $b->{start}) || ($a->{strand} <=> $b->{strand}) || ($a->{end} <=> $b->{end});
	}
	@{$record->{gene_list}} = sort by_position @Gene_List;
	$self->{record_hash}->{$fragment_accession} = $record if (defined $self->{record_hash});
	
	#Create a cache file
	if ($cache_file)
	{
		Storable::store($record, $cache_file);
	#	open CACHE, ">$cache_file";
	#	print CACHE unpack( "H*", Storable::nfreeze($record) ) . "\n";
	#	close CACHE;
	}
	
	return $record;
}


sub GetTag
{
	my ($Feature, $Tag, $Allow_Array) = @_;
	$Allow_Array = 0 if (!defined $Allow_Array);
	
	return '' if (!$Feature->has_tag("$Tag"));
	my @Tag_Values = $Feature->get_tag_values("$Tag");
	return '' if (scalar @Tag_Values == 0);
	return $Tag_Values[0] if (!$Allow_Array);
	return @Tag_Values;
}	

sub CalculateLength
{
	my ($In_Start, $In_End, $In_Sequence_Length, $In_Topology) = @_;

	#Start will always be less than end, except when we are overlapping the end of a circular genome
	
	if ($In_End < $In_Start)
	{
		if ($In_Topology eq 'circular')
		{
			$In_End += $In_Sequence_Length;
		}
		else #Linear
		{
			return undef;
		}
	}
	
	my $Length = $In_End - $In_Start + 1;
	
	
	return $Length;
}

sub OffsetPosition
{
	my ($In_Start, $In_Offset, $In_Sequence_Length, $In_Topology) = @_;
	my $Out_Start = $In_Start + $In_Offset;
	
	
	if ($Out_Start < 1)
	{
		$Out_Start = ($In_Topology eq 'circular') ? $Out_Start + $In_Sequence_Length : 1 ;
	}
	
	if ($Out_Start > $In_Sequence_Length)
	{
		$Out_Start = ($In_Topology eq 'circular') ? $Out_Start - $In_Sequence_Length : $In_Sequence_Length;
	}
	
	return $Out_Start;
}

return 1;


