###
# Pod Documentation
###

=head1 NAME

GenomeContextDrawer.pm

=head1 SYNOPSIS

=head1 DESCRIPTION

Object that draws genome contexts of hits.

=head1 AUTHOR

Jeffrey Barrick

=head1 COPYRIGHT

Copyright (C) 2008 Jeffrey Barrick.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut

###
# End Pod Documentation
###


use GD;

use Bio::Graphics;
use Bio::SeqFeature::Generic;
use Bio::SeqIO;

use GenomeContext;

package GenomeContextDrawer;
use vars qw(@ISA);
use strict;

@ISA = qw( Bio::Root::RootI );


=head2 new

 Title   : new
 Usage   : $dbcache = Intergenic::Cache->new( -name => 'cog');
 Function: Creates/opens a database to cache values to
 Returns : a new DatabaseCache object

=cut

sub new
{
	my($caller,@args) = @_;
	my $class = ref($caller) || $caller;
	my $self = Bio::Root::Root->new($caller, @args);
	bless ($self, $class);
	
	($self->{svg}, $self->{cdd_desc}) = $self->Bio::Root::Root::_rearrange([qw(SVG CDD_DESC)], @args);
	
	if (!$self->{cdd_desc})
	{
		print "No CDD info table loaded. Drawings will not have CDD gene names.\n";
	}
	
	return $self;
}

our @Color_List = ('blue', 'green', 'red', 'orange', 'purple', 'yellow', 'aqua', 'brown', 'magenta', 'olive', 'chartreuse', 
     'darkgoldenrod', 'darksalmon', 'darkorchid', 'cornflowerblue', 'maroon', 'mediumspringgreen', 'forestgreen', 'darkkhaki', 'navy', 'salmon', 
     'indigo', 'burlywood', 'crimson', 'chocolate', 'lavender', 'lime', 'olive', 'plum', 'seagreen', 'teal', 'wheat', 
     'orangered', 'coral', 'lightcyan', 'greenyellow', 'black', 'gray' );

#Create a drawing for each prediction
sub draw
{
	my ($self, %options) = @_;
	#use Data::Dumper;
	#print Dumper(\%options);
		
	$self->throw("Must provide option 'out_path'") if (!defined $options{out_path});
	mkdir $options{out_path} if (!-e $options{out_path});

	#Load all of the feature lists (so we can decide to color cogs intelligently as a group)	
	$self->throw("Must provide option 'regions' = handle to array of GenomeContext objects") if (!defined $options{regions});
	
	my $feature_list = $options{regions};

	#Calculate color table based on COGs
	our %CDD_Color_Hash = ();
	my %CDD_Hash = ();
	
	foreach my $features (@$feature_list)
	{
		foreach my $gene (@{$features->{neighborhood}})
		{	
			foreach my $cdd (keys %{$gene->{cdd_hash}})
			{
				$CDD_Hash{$cdd}++;
			}
		} 
	}
	
	my $count = 0;
	foreach my $key (sort { (-($CDD_Hash{$a} <=> $CDD_Hash{$b})) || ($a cmp $b) } keys %CDD_Hash)
	{
		#If the cog only occurs once, then make it gray
		$CDD_Color_Hash{$key} = $Color_List[$count];
		$count++ if ($count < $#Color_List);
		$CDD_Color_Hash{$key} = 'white' if ($CDD_Hash{$key} == 1);
	}
	
	##
	# Start drawing
	##	
	
	my $i = 0;
	my ($w, $h);
	my $out_file_name;

	NBHD: foreach our $features (@$feature_list)
	{
		#Initialize certain variables in case we bail midway
		($w, $h) = (0, 0);
		$out_file_name = '';
		
		$i++;
		
		#This neighborhood was not loaded correctly.
		our $pad_right = 0;
		our $width = 800;
		
		my $Panel = Bio::Graphics::Panel->new(-image_class=>($self->{svg} ? 'GD::SVG' : 'GD'), -start => $features->{interval}->{start}, -end => $features->{interval}->{end}, -width => $width, -pad_left => 60, -pad_right => $pad_right + 60, -pad_top => 3, -pad_bottom => 3, -flip => ($features->{interval}->{strand} == -1));
		
		$features->{interval}->{length} = $features->{interval}->{end} - $features->{interval}->{start} + 1;
		
		our $display = {};
		$display->{start} = ($features->{main}->{strand} == +1) ?  $features->{main}->{start} : $features->{main}->{end};
		$display->{end}   = ($features->{main}->{strand} == +1) ?  $features->{main}->{end} : $features->{main}->{start};
		
		##Option shared by several tracks
		my $Glyph_Hash = { match => 'transcript2', cds => 'transcript', cdd => 'box' };

		##Create the gene track
		my $Track = $Panel->add_track(-glyph => $Glyph_Hash, -bump => -1, -label => 1, -bgcolor => \&GetGeneColor, -fontcolor => 'blue', -font2color => 'green', -height=>8, -font => 'gdTinyFont', -description => \&GetDescription);

		##Create the ruler
		my $Glyph_Hash_1 = { match => 'transcript2', ruler => 'arrow' };
		my $Ruler_Track = $Panel->add_track(-label => 1, -glyph => $Glyph_Hash_1, -tick => 2, -fgcolor => 'black', -bgcolor => 'black', -bump => 0);

		my $Ruler = Bio::SeqFeature::Generic->new(-display_name => "$features->{fragment_accession}/$display->{start}-$display->{end} $features->{description}", -primary_tag => 'ruler', -start => $features->{interval}->{start}, end => $features->{interval}->{end} );
		$Ruler_Track->add_feature($Ruler);	
		
		##Create the main "hit" object 
		my $Match = Bio::SeqFeature::Generic->new(-primary_tag => 'match', -display_name => ' ', -start => $features->{main}->{start}, -end => $features->{main}->{end}, -strand => $features->{main}->{strand} );
		$Ruler_Track->add_feature($Match);

		##Create the CDD track
		my $CDD_Track = $Panel->add_track(-glyph => $Glyph_Hash, -bump => +1, -label => 1, -bgcolor => \&GetColor, -strand_arrow  => 0, -fontcolor => 'red', -font2color => 'red', -height=>8, -font => 'gdTinyFont', -description => 0);

		sub GetGeneColor 
		{
			my $Feature = shift;  
			if ( $Feature->has_tag('-bgcolor') )
			{
				my @Tag_List = $Feature->get_tag_values('-bgcolor');
				return $Tag_List[0];
			}
			return 'white';
		}

		sub GetColor 
		{ 
			my $Feature = shift;  
				
			return 'black' if ($Feature->primary_tag eq 'match');

			if ( $Feature->has_tag('cdd') )
			{
				my @Tag_List = $Feature->get_tag_values('cdd');
				if ($CDD_Color_Hash{$Tag_List[0]})
				{
					return $CDD_Color_Hash{$Tag_List[0]};
				}
			}

			return 'white';
		}
		
		sub GetDescription
		{
			my $feature = shift;
			my $desc;
			 
			#Create the ideal description if we had infinite space 
		#	return '' if ( !$feature->has_tag('cdd') and !$feature->has_tag('desc') );
		#	my @Tag_List = $feature->get_tag_values('cdd');
		#	$desc = $Tag_List[0];
		#	$desc .= " " if ($desc);
			return '' if (!$feature->has_tag('desc'));  #alternate
			
			my @Tag_List = $feature->get_tag_values('desc');
			#Remove "COGXXXX: " from beginning of some descriptions since it is redundant
			#$Tag_List[0] =~ s/^COG....: //;
			$desc .= $Tag_List[0];
			
			#Calculate how much space in pixels we have based on our own left position, then add right margin
			my $allowed_pixels;
			if ($features->{interval}->{strand} == +1)
			{
				$allowed_pixels =  (($features->{interval}->{end} - $feature->start) / $features->{interval}->{length} ) * $width + $pad_right;
			}
			else
			{
				$allowed_pixels = (($feature->end - $features->{interval}->{start}) / $features->{interval}->{length} ) * $width + $pad_right;
			}
			
			$allowed_pixels -= GD::gdTinyFont->width * (length '...');
			my $allowed_char =  int ($allowed_pixels / GD::gdTinyFont->width);
			
			#Add letters until we're out of space.	
			$desc = (substr $desc, 0, $allowed_char) . '...' if (length $desc > $allowed_char) ;
		
			return $desc;
			
		}

		#All Genes		
		foreach our $Gene (@{$features->{neighborhood}})
		{	
			
			#Adjust the bounds of features that go past the ends of the interval for correct printing
			$Gene->{start} = $features->{interval}->{start} if ($Gene->{start} < $features->{interval}->{start});
			$Gene->{end}   = $features->{interval}->{end}   if ($Gene->{end}   > $features->{interval}->{end}  );
			$Gene->{name} = 'unknown' if (!$Gene->{name});			

			#Add accesion to name
			$Gene->{name} .= " $Gene->{accession}";


			sub by_domain
			{
				#Find the minimum start for each CDD being compared
				my $a_max;
				my $a_min;
				foreach my $cdd (@{$Gene->{cdd_hash}->{$a}})
				{
					$a_min = $cdd->{start} if (!defined $a_min or $cdd->{start} < $a_min);
					$a_max = $cdd->{end} if (!defined $a_max or $cdd->{end} > $a_max);
				}
				
				my $b_max;
				my $b_min;
				foreach my $cdd (@{$Gene->{cdd_hash}->{$a}})
				{
					$b_min = $cdd->{start} if (!defined $b_min or $cdd->{start} < $b_min);
					$b_max = $cdd->{end} if (!defined $b_max or $cdd->{end} > $b_max);
				}
				
				return ($Gene->{cdd_hash}->{$a}->[0]->{strand} == +1) ? 
					($a_min < $b_min) :
					($a_max > $b_max) ;
			}
			my @cdd_list = sort by_domain keys %{$Gene->{cdd_hash}};
			my $cdd_string = join "/", @cdd_list;
			#Split genes that are in multiple cogs
		#	if (scalar @cdd_list > 1)
			{
				my $feature = Bio::SeqFeature::Generic->new(-primary_tag => 'cds', -display_name => "$Gene->{name}" , -start => $Gene->{start}, -end => $Gene->{end}, -strand => $Gene->{strand}, -tag => { cdd => $cdd_string, desc => $Gene->{description}, -bgcolor => 'white'} );	
				my @domains;
				
				for (my $i=0; $i < scalar @cdd_list; $i++)
				{					
					my $key = $cdd_list[$i];
					my $cdd_display_name = $key;
					if ($self->{cdd_desc})
					{
						my $info = $self->{cdd_desc}->cdd_info($key);
						$cdd_display_name .= " $info->{gene}";
						
					}
					CDD: foreach my $cdd (@{$Gene->{cdd_hash}->{$key}})
					{
						my $cdd_feature = Bio::SeqFeature::Generic->new(-primary_tag => 'cdd', -display_name => "$cdd_display_name" , -tag => { cdd => $key }, -start => $cdd->{start} , -end => $cdd->{end}, -strand => 0);
						$CDD_Track->add_feature($cdd_feature);	
					
						my $new_domain = { start => $cdd->{start}, end =>  $cdd->{end} };

						#See if we overlap any current domains (in which case we expand their boundaries and possibly join them)
						my @join_list;
						for (my $i=0; $i < scalar @domains; $i++)
						{
							my $domain = $domains[$i];
							
							if ( (($domain->{start} <=  $cdd->{start}) && ($domain->{end} >= $cdd->{start}))
							||   (($domain->{start} <=  $cdd->{end}) && ($domain->{start} >= $cdd->{start})) )
							{
								push @join_list, $i;
								#Now see if this runs into any other domains
							}
						}
						
						if (scalar @join_list > 0)
						{
							foreach my $j (reverse @join_list)
							{
								my $domain = $domains[$j];
								$new_domain->{start} = ($new_domain->{start} < $domain->{start}) ? $new_domain->{start} : $domain->{start};
								$new_domain->{end} = ($new_domain->{end} > $domain->{end}) ? $new_domain->{end} : $domain->{end};
								
								#Remove the overlapped
								splice @domains , $j, 1;
							}							
						}

						push @domains, $new_domain;
					}
					
				}
				
				my $on_start = $Gene->{start};
				
				@domains = sort {$a->{start} <=> $b->{start}} @domains;
				foreach my $domain (@domains)
				{
					#Non-CDD before
					my $before_feature = Bio::SeqFeature::Generic->new(-primary_tag => 'cds', -display_name => "$Gene->{name}" , -tag => { desc => '', -bgcolor =>'white'}, -start => $on_start , -end => $domain->{start}-1, -strand => $Gene->{strand});
					$feature->add_sub_SeqFeature($before_feature,'EXPAND');
					$on_start = $domain->{end} + 1;
					
					my $sub_feature = Bio::SeqFeature::Generic->new(-primary_tag => 'cds', -display_name => "$Gene->{name}" , -tag => { desc => '', -bgcolor =>'black'}, -start => $domain->{start} , -end => $domain->{end}, -strand => $Gene->{strand});
					$feature->add_sub_SeqFeature($sub_feature,'EXPAND');
				}
				my $after_feature = Bio::SeqFeature::Generic->new(-primary_tag => 'cds', -display_name => "$Gene->{name}" , -tag => { desc => '', -bgcolor =>'white'}, -start => $on_start , -end => $Gene->{end}, -strand => $Gene->{strand});
				$feature->add_sub_SeqFeature($after_feature,'EXPAND');
				
				$Track->add_feature($feature);
			}			
	#		else
	#		{
	#			$Gene->{cdd} = join "|", (sort {$Gene->{cdd_hash}->{$a} <=> $Gene->{cdd_hash}->{$b}} keys %{$Gene->{cdd_hash}});
	#			my $feature = Bio::SeqFeature::Generic->new(-primary_tag => 'cds', -display_name => "$Gene->{name}" , -start => $Gene->{start}, -end => $Gene->{end}, -strand => $Gene->{strand}, -tag => { cdd => $cdd_string, desc => ''} );		
	#			$Track->add_feature($feature);
	#		}
		}
		
		#Create the CDD object
		
		$out_file_name = ($self->{svg} ? "$i.svg" : "$i.png");
							
		open DRAWING, ">$options{out_path}/$out_file_name"  or die "Could not open output file $options{out_path}/$out_file_name";
		my $gd = $Panel->gd; 
		print DRAWING $gd->png if (!$self->{svg});
		print DRAWING $gd->svg if ($self->{svg});
		close DRAWING;
		($w, $h) = $gd->getBounds();
		
		#Call or else there is a memory leak from circular references
		$Panel->finished;

	} continue {

		#Save important information about what we have drawn
		my $ai = $i-1;
		$self->{output}->[$ai] = { name=> $out_file_name, height=>$h, width=>$w };		
		if (defined $options{relative_url})
		{	
			if ($self->{svg})
			{
				$self->{output}->[$ai]->{html} =  "<embed src=\"$options{relative_url}/$out_file_name\" width=$w height=$h pluginspage=\"http://www.adobe.com/svg/viewer/install/\">";
			}
			else
			{
				$self->{output}->[$ai]->{html} =  "<img src=\"$options{relative_url}/$out_file_name\">";
			}
		}
	}
}



