#!/usr/bin/perl -w

###
# Pod Documentation
###

=head1 NAME

extend_stockholm.perl

=head1 SYNOPSIS

Usage: decorate_aligment.perl -i <input> -o <output> [-c taxonomy_file] [-g organism] [-e #nt]

Add columns to an alignment

=head1 DESCRIPTION

=over

=item B<-i> <file path> 

Input Stockholm file.

=item B<-o> <file path> 

Output Stockholm file.

=item B<-d> <file path>

Path to the index for a databasecontaining the full sequences of the hits in the alignment.

=item B<-e> <#nt> 

Extend the alignment by this many nt on both sides.

=item B<-1> <#nt> 

Extend the alignment by this many nt upstream.

=item B<-2> <#nt> 

Extend the alignment by this many nt downstream.

=back

=head1 AUTHOR

Jeffrey Barrick

=head1 COPYRIGHT

Copyright (C) 2008 Jeffrey Barrick.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut

###
# End Pod Documentation
###


use strict;
use Data::Dumper;

use FindBin;
use lib $FindBin::Bin;

use Bio::DB::Flat;
use Bio::AlignIO;

#Get options
use Getopt::Long;
use Pod::Usage;
my ($help, $man);
my ($db_path, $extend, $extend_upstream, $extend_downstream, $input_file, $output_file);
$extend = 0;
$extend_upstream = 0;
$extend_downstream = 0;
#pod2usage(1) if (scalar @ARGV == 0);
GetOptions(
	'help|?' => \$help, 'man' => \$man,
	'extend|e=s' => \$extend,
	'upstream|1=s' => \$extend_upstream,
	'downstream|2=s' => \$extend_downstream,
	'in|i=s' => \$input_file,
	'out|o=s' => \$output_file,
	'database-path|d=s' => \$db_path,
) or pod2usage(2);
pod2usage(1) if $help;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;
pod2usage(1) if (!defined $input_file || !defined $output_file || !defined $db_path);

$extend_upstream += $extend;
$extend_downstream += $extend;

#Open database
$db_path =~ m/^(.+)\/(.+)$/;
my ($db_dir, $db_name) = ($1, $2);

#Open Input
my $in = Bio::AlignIO->new(-format => 'stockholm', -file   => $input_file);
my $aln = $in->next_aln();

#Open output
my $out = Bio::AlignIO->new(-format => 'stockholm', -file   => ">$output_file");
$out->line_length(0);

my @new_seqs;
foreach my $seq ( $aln->each_seq() )
{
	push @new_seqs, $seq;
	$aln->remove_seq($seq);

    #print +scalar @new_seqs . " $name\n";	

    my $accession = $seq->display_name;
	my $start = $seq->start;
	my $end = $seq->end;
	my $strand = ($start > $end) ? -1 : 1;
	my $extend_sequence = '';
	
	#Extend
	print +(scalar @new_seqs) . ": Loading $accession\n";
	
	#For some reason the database goes dead after one read...
	my $db = Bio::DB::Flat->new(	-directory  => $db_dir,
			                        -dbname     => $db_name,
			                        -format     => 'genbank',
			                        -index      => 'flat',
			                        -write_flag => 0);
		                        
	my $full_seq = $db->get_Seq_by_id($accession);
	print $full_seq->accession . "\n";	
	
	my $extend_end;
	my $extend_start;
	my $limit_extend_end;
	my $limit_extend_start;
		
	if ($strand == +1)
	{
		$extend_end = $end + $extend_downstream;
		$extend_start = $start - $extend_upstream;
				
		$limit_extend_end = ($extend_end > $full_seq->length) ? $full_seq->length : $extend_end;
		$limit_extend_start = ($extend_start < 1) ? 1 : $extend_start;
		
		$extend_sequence .= '.' x ($limit_extend_start - $extend_start);
		if ($start != $limit_extend_start)
		{
			my $add_seq = $full_seq->trunc($limit_extend_start, $start - 1);
			$extend_sequence .= $add_seq->seq;
		}
		
		$extend_sequence .= $seq->seq;
		
		$extend_sequence .= '.' x ($extend_end - $limit_extend_end);
		if ($end != $limit_extend_end)
		{
			my $add_seq = $full_seq->trunc($end+1, $limit_extend_end);
			$extend_sequence .= $add_seq->seq;
		}
	}
	else
	{
		$extend_end = $end - $extend_downstream;
		$extend_start = $start + $extend_upstream;
		
		$limit_extend_start = ($extend_start > $full_seq->length) ? $full_seq->length : $extend_start;
		$limit_extend_end = ($extend_end < 1) ? 1 : $extend_end;
		
		$extend_sequence .= '.' x ($extend_start - $limit_extend_start);
		if ($start != $limit_extend_start)
		{
			my $add_seq = $full_seq->trunc($start + 1, $limit_extend_start);
			$add_seq = $add_seq->revcom();
			$extend_sequence .= $add_seq->seq;
		}
		
		$extend_sequence .= $seq->seq;
		
		$extend_sequence .= '.' x ($limit_extend_end - $extend_end);
		if ($end != $limit_extend_end)
		{
			my $add_seq = $full_seq->trunc($limit_extend_end, $end-1);
			$add_seq = $add_seq->revcom();
			$extend_sequence .= $add_seq->seq;
		}
	}
	
	print "$accession $start $end $strand => $extend_start $extend_end $strand\n";

	$extend_sequence =~ tr/T/U/;

	print $seq->seq. "\n" . "$extend_sequence\n";

	$seq->start($limit_extend_start);
	$seq->end($limit_extend_end);
	$seq->seq($extend_sequence);
}
	
	#Also extend all GC lines

my $ameta = $aln->consensus_meta;
foreach my $mname ($ameta->meta_names)
{
	my $replace_char = '.';
	if ($mname eq 'SS_cons')
	{
		$replace_char = ':'; #proper symbol for outside of all pairing...
	}	
	
	$ameta->named_meta($mname, ($replace_char x $extend_upstream). $ameta->named_meta($mname) . ($replace_char x $extend_downstream));
}
	
#	foreach my $key (keys %{$sto->{stockholm}->{GC}})
#	{
#	    my $ch=".";
#	    if ($key eq "HYPHENATION") {
#		$ch="*";
#	    }
#	    $sto->{stockholm}->{GC}->{$key} = ($ch x $extend) . $sto->{stockholm}->{GC}->{$key} . ($ch x $extend);
#	}
#
#	push @new_seqs, $seq;


foreach my $seq (@new_seqs)
{
	$aln->add_seq($seq);
}
$out->write_aln($aln);

