#!/usr/bin/perl -w

###
# Pod Documentation
###

=head1 NAME

fragment_db.pl

=head1 SYNOPSIS

Usage: fragment_db.pl -i <input> -o <output> [-c taxonomy_file] [-g organism] [-e #nt]

Split a directory of sequence files

=head1 DESCRIPTION

=over

=item B<-s> <#megabytes> 

Minimum size of sequence file to create (in megabytes).

=item B<-p> <regexp pattern> 

Only include sequence files that match this pattern in the input directory.

=item B<-r> <file prefix>

Add this prefix to the fragment files in the output directory.

=back

=head1 AUTHOR

Jeffrey Barrick

=head1 COPYRIGHT

Copyright (C) 2008 Jeffrey Barrick.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut

###
# End Pod Documentation
###


use strict;
use Getopt::Long;
use Pod::Usage;

my ($help, $man);
my ($opt_p, $opt_s, $opt_r);	
$opt_s = 10; #default 10 megabytes

GetOptions(
	'help|?' => \$help, 'man' => \$man,
	'size|s=s' => \$opt_s,
	'pattern|p=s' => \$opt_p,
	'prefix|2=s' => \$opt_r,
) or pod2usage(2);
pod2usage(1) if $help;
pod2usage(-exitstatus => 0, -verbose => 2) if $man;
my ($opt_i, $opt_o) = @ARGV;
pod2usage(1) if (!defined $opt_i || !defined $opt_o);

my $fragment_size = $opt_s * 1000000;

opendir(INDIR, $opt_i) or die "Cannot open input directory: $opt_i";
my @In_Files = grep !/^\./, readdir INDIR;
@In_Files = grep /$opt_p/, @In_Files if ($opt_p);

die "No files in list!" if (scalar @In_Files == 0);

use Bio::SeqIO;

my $chunk_index = 1;

mkdir $opt_o if (! -e $opt_o);

my $file_name = "$opt_o/";
$file_name .= "$opt_r\_" if ($opt_r);
$file_name .= "$chunk_index.fna";

my $seq_out = Bio::SeqIO->new(-file => ">$file_name");
foreach my $file (@In_Files)
{
	my $seq_in = Bio::SeqIO->new(-file => "<$opt_i/$file", -format => 'fasta');
	while (my $seq = $seq_in->next_seq())
	{
		$seq_out->write_seq($seq);
		
		#Open a new output file once a certain size is exceeded
		if (-s "$file_name" > $fragment_size)
		{
			$chunk_index++;
			$file_name = "$opt_o/";
			$file_name .= "$opt_r\_" if ($opt_r);
			$file_name .= "$chunk_index.fna";
			$seq_out = Bio::SeqIO->new(-file => ">$file_name", -format => 'fasta');
		}
	}		
}